REM
REM --------------------------------------
REM This file is part of the zBoy project.
REM Copyright (C) Mateusz Viste 2010
REM --------------------------------------
REM
REM * MBC1 short documentation *
REM
REM MBC1 can operate in two modes:
REM 16Mbit ROM / 8 KByte RAM (default at power on)
REM 4Mbit ROM / 32 KByte RAM
REM
REM Write a byte into 6000-7FFF to select the mode:
REM Any EVEN byte selects 16/8 mode, any UNEVEN byte selects 4/32 mode (only the last bit is looked at).
REM
REM Write a byte into 2000-3FFF to select the ROM bank for 4000-7FFF (bank #0 will point to bank #1 anyway)
REM Only 5 least significant bits are looked at (xxxBBBBB)
REM
REM For mode 4/32, you need to select the RAM bank to be used (at A000-C000): to do that, write a byte into
REM 4000-5FFF. The 2 least significant bits of this byte (xxxxxxBB) are telling what RAM bank to use.
REM Also, before using any RAM bank, you need to enable it by writing xxxx1010 into 0000-1FFF (writing there
REM any non-xxxx1010 value disables the RAM bank).
REM
REM For mode 16/8, 5 bits is not enough to address all 128 banks (each bank is 16KiB big). Write a byte into
REM 4000-5FFF. The 2 least significant bits of this byte (xxxxxxBB) will be used as the two most significant
REM bits (that is bits 5 and 6) of the ROM address.
REM
REM Note: The content of the MBC1 RAM can be saved during poweroff if the cartridge comes with a battery.
REM


SUB MBC1_MemoryRead(BYREF memaddr AS UINTEGER, BYREF Result AS UBYTE)
  IF (memaddr <= &h00FF) AND (BootRomEnabledFlag = 1) THEN
    Result = BootROM(memaddr)
  ELSEIF (memaddr >= &hA000) AND (memaddr < &hC000) THEN        ' RAM bank n
    IF (Mbc1Model = 2) THEN   ' 4/32 mode
        'Result = MemoryBankedRAM((CurRamBank SHL 13) + (memaddr - &hA000))     ' SHL 13 is the same than *8192 (but faster)
        Result = MemoryBankedRAM((Mbc1RamRomSelect SHL 13) + (memaddr))     ' SHL 13 is the same than *8192 (but faster)
      ELSE   ' 16/8 mode
        Result = MemoryBankedRAM(memaddr)
    END IF
  ELSEIF (memaddr >= &hC000) AND (memaddr < &hE000) THEN    ' Internal 8KiB RAM
    Result = MemoryInternalRAM(memaddr)
  ELSEIF (memaddr >= &hE000) AND (memaddr < &hFE00) THEN    ' RAM mirror
    Result = MemoryInternalRAM(memaddr - 8192)
  ELSEIF (memaddr >= &hFE00) AND (memaddr < &hFEA0) THEN    ' Sprite OAM memory
    Result = SpriteOAM(memaddr)
  ELSEIF (memaddr >= &hFF80) AND (memaddr <= &hFFFF) THEN   ' Hi RAM area
    Result = MemoryInternalHiRAM(memaddr)
  ELSEIF (memaddr >= &h0000) AND (memaddr < &h4000) THEN    ' ROM bank #0
    Result = MemoryROM(memaddr)
  ELSEIF (memaddr >= &hFF00) AND (memaddr <= &hFF4B) THEN   ' I/O registers
    Result = IoRegisters(memaddr)
  ELSEIF (memaddr >= &h4000) AND (memaddr < &h8000) THEN    ' ROM bank #n
    'IF RomInfos.RomSize = 32768 THEN
    '    RETURN MemoryROM(memaddr)
    '  ELSE
        'IF (CurRomBank = 0) OR (RomInfos.RomSize = 32768) THEN
        IF (RomInfos.RomSize = 32768) THEN
            'RETURN MemoryROM(16384 + (memaddr - &h4000))
            Result = MemoryROM(memaddr)
          ELSE
            IF Mbc1Model = 2 THEN    '  4/32
                IF ((CurRomBank SHL 14) + (memaddr - &h4000)) > RomInfos.RomSize THEN                  ' SHL 14 is the same than *16384 (but faster)
                  PrintMsg("INVALID ROM ADDRESS!" & CHR(10) & "CURBANK=" & CurRomBank & CHR(10) & "MBC1 MODEL=" & Mbc1Model)
                  PressAnyKey()
                  END
                END IF
                Result = MemoryROM((CurRomBank SHL 14) + (memaddr - &h4000))    ' SHL 14 is the same than *16384 (but faster)
              ELSE   ' 16/8
                IF (((CurRomBank OR (Mbc1RamRomSelect SHL 5)) SHL 14) + (memaddr - &h4000)) > RomInfos.RomSize THEN                  ' SHL 14 is the same than *16384 (but faster)
                  PrintMsg("INVALID ROM ADDRESS!" & CHR(10) & "CURBANK=" & CurRomBank & CHR(10) & "MBC1 MODEL=" & Mbc1Model)
                  PressAnyKey()
                  END
                END IF
                Result = MemoryROM(((CurRomBank OR (Mbc1RamRomSelect SHL 5)) SHL 14) + (memaddr - &h4000))    ' SHL 14 is the same than *16384 (but faster)
            END IF
        END IF
    'END IF
  ELSEIF (memaddr >= &h8000) AND (memaddr <= &h9FFF) THEN   ' Video RAM (8KiB)
    Result = VideoRAM(memaddr)
  ELSE                                                  ' Else it's something I don't know about
    'IF (memaddr > &hFFFF) THEN
    '  PrintMsg("MMU FATAL ERROR!" & CHR(10) & "MEMORYREAD() -> INVALID MEMADDR! [" & HEX(memaddr) & "]") : PressAnyKey() : END
    'END IF
    'PRINT "MMU fatal error: Tried to read from an unknown memory address: " & HEX(memaddr) & "h"
    'SLEEP
    'QuitEmulator = 1
    Result = MemoryMAP(memaddr)
  END IF
END SUB


SUB MBC1_MemoryWrite(BYREF memaddr AS UINTEGER, BYREF DataByte AS UBYTE)
  IF (memaddr < &h2000) THEN  ' Enables / disables current RAM bank (not supported yet)
    REM Before using any RAM bank, you need to enable it by writing xxxx1010 into 0000-1FFF (writing there
    REM any non-xxxx1010 value disables the RAM bank).
    'IF (DataByte AND &b00001111) = &b00001010 THEN ' Enables currently selected RAM bank
    '  ELSE  ' Disables currently selected RAM bank
    '    'UserMessage = "RAM BANK DISABLED"
    'END IF
  ELSEIF (memaddr >= &h2000) AND (memaddr < &h4000) THEN       ' Select current ROM or RAM bank on MBC1
    REM TODO checklist #784632
    'IF (Mbc1Model = 1) THEN  ' Mode 16/8
    '    CurRomBank = (CurRomBank AND &b01100000) OR (DataByte AND &b00011111)
        'CurRomBank = (DataByte AND &b00011111)
    '  ELSE  ' Mode 4/32
        CurRomBank = (DataByte AND &b00011111)
    'END IF
    IF CurRomBank = &h00 THEN CurRomBank = &h01  ' Should I do that? Not super sure...
  ELSEIF (memaddr >= &h4000) AND (memaddr < &h6000) THEN
    Mbc1RamRomSelect = (DataByte AND &b00000011)
    'SELECT CASE MbcModel ' For mode 16/8, 5 bits is not enough to address all 128 banks (each bank is 16KiB big). Write a
    '  CASE 1 ' byte into 4000-5FFF. The 2 least significant bits of this byte (xxxxxxBB) will be used as the two most
    '    CurRomBank = (CurRomBank AND &b00011111) OR ((DataByte SHL 5) AND &b01100000)  '  significant bits of the ROM address.
    '  CASE 2 ' For mode 4/32, you need to select the RAM bank to be used (at A000-C000). to do that, write a
    '    CurRamBank = (DataByte AND &b00000011)    ' byte into 4000-5FFF. The 2 least significant bits of this
    'END SELECT                                     ' byte (xxxxxxBB) are telling what RAM bank to use.
  ELSEIF (memaddr >= &h6000) AND (memaddr < &h8000) THEN   ' Configuration for the MBC1 chip (selection of mem model)
    IF (DataByte AND &b00000001) = 0 THEN
        Mbc1Model = 1   ' 16/8 model
      ELSE
        Mbc1Model = 2   ' 4/32 model
    END IF
  ELSEIF (memaddr >= &hA000) AND (memaddr < &hC000) THEN    ' RAM bank n
    IF Mbc1Model = 2 THEN   ' 4/32 mode
        'MemoryBankedRAM((CurRamBank SHL 13) + (memaddr - &hA000)) = DataByte     ' SHL 13 is the same than * 8192 (but faster)
        MemoryBankedRAM((Mbc1RamRomSelect SHL 13) + memaddr) = DataByte     ' SHL 13 is the same than * 8192 (but faster)
      ELSE   ' 16/8 mode
        MemoryBankedRAM(memaddr) = DataByte
    END IF
  ELSEIF (memaddr >= &hC000) AND (memaddr < &hE000) THEN    ' Internal 8KiB RAM
    MemoryInternalRAM(memaddr) = DataByte
  ELSEIF (memaddr >= &hE000) AND (memaddr < &hFE00) THEN    ' RAM mirror
    MemoryInternalRAM(memaddr - 8192) = DataByte
  ELSEIF (memaddr >= &hFE00) AND (memaddr < &hFEA0) THEN    ' Sprite OAM memory
    SpriteOAM(memaddr) = DataByte
  ELSEIF (memaddr >= &hFF80) AND (memaddr <= &hFFFF) THEN   ' Hi RAM area
    MemoryInternalHiRAM(memaddr) = DataByte
  ELSEIF (memaddr = &hFF41) THEN ' STAT register: Do not allow to write into 2 last bits of the STAT register, as these
    IoRegisters(&hFF41) = (IoRegisters(&hFF41) AND &b00000011) OR (DataByte AND &b11111100)    ' bits are the mode flag.
  ELSEIF (memaddr = &hFF44) THEN ' CURLINE [RW] Current Scanline.
    IoRegisters(memaddr) = 0     ' Writing into this register resets it.
    'UserMessage = "LY RESET"
  ELSEIF (memaddr = &hFF46) THEN ' Starts LCD OAM DMA transfer
    STATIC x AS INTEGER          '
    IF DataByte <= &hF1 THEN     ' Normally, should take 160us... but who cares?
        FOR x = 0 TO 159         ' Let's copy XX00-XX9F to FE00-FE9F
          MBC1_MemoryRead((DataByte SHL 8) + x, SpriteOAM(&hFE00 + x))
        NEXT x
      ELSE
        PrintMsg("MMU FATAL ERROR!" & CHR(10) & "TRIED TO RUN LCD OAM DMA TRANSFER USING AN INVALID RANGE!")
        PressAnyKey()
        END
    END IF
  ELSEIF(memaddr = &hFF04) THEN
    IoRegisters(&hFF04) = 0      ' Divide register: Writing any value sets it to 0
  ELSEIF (memaddr >= &hFF00) AND (memaddr <= &hFF4B) THEN   ' I/O registers
    IoRegisters(memaddr) = DataByte
  ELSEIF (memaddr = &hFF50) THEN
    BootRomEnabledFlag = 0
  ELSEIF (memaddr >= &h8000) AND (memaddr <= &h9FFF) THEN   ' Video RAM (8KiB)
    VideoRAM(memaddr) = DataByte
  'ELSEIF (memaddr >= &hFEA0) AND (memaddr <= &hFEFF) THEN   ' This area is not usable, but some games writes to it...
  '  REM                                                     ' For example Dr. Mario tries to write to FEFF... who knows why?
  'ELSEIF (memaddr >= &hFF4C) AND (memaddr <= &hFF7F) THEN   ' This area is not usable, but some games writes to it...
  '  REM                                                     ' For example Dr. Mario tries to write to FF7F... who knows why?
  ELSE                                                      ' Else it's something I don't know about
    'PRINT "MMU fatal error: Tried to write to an unknown memory address: "; HEX(memaddr)
    'QuitEmulator = 1
    MemoryMAP(memaddr) = DataByte
  END IF
END SUB
